require 'delegate'

module SaveTheMonth
  class Movement < Struct.new(:type, :name, :amount)
    def can_perform?(given_date, initial_date)
      true
    end
  end
  
  class DailyRestrictedMovement < SimpleDelegator
    attr_accessor :date
    
    def can_perform?(given_date, initial_date)
      given_date == date
    end
  end
  
  class PeriodicalMovement < SimpleDelegator
    attr_accessor :can_perform_strategy
    
    def can_perform?(given_date, initial_date)
      can_perform_strategy.call(given_date, initial_date)
    end
  end
  
  PeriodicalStrategy = Struct.new(:amount, :can_perform)
  
  class PaymentPeriodicalStrategies
    def daily(amount)
      # everyday this should be substracted or added
      can_perform = ->(given_date, initial_date){ true }
      PeriodicalStrategy.new(amount, can_perform)
    end
    
    def weekly(amount)
      can_perform = ->(given_date, initial_date) do
        # every week. I can only happen if today
        # is one of the next days: [6, 13, 20 or 27 days after the initial date]
        week_days = [ initial_date + 6, initial_date + 13, initial_date + 20, initial_date + 27]
        week_days.include?(given_date)
      end
      
      PeriodicalStrategy.new(amount, can_perform)
    end
  end
  
  class ExpectedBalance
    attr_accessor :initial_date, :final_date, :initial_amount, :movements, :periodical_strategies
    
    def initialize
      self.movements = []
      self.periodical_strategies = PaymentPeriodicalStrategies.new
    end
    
    def add_charge(name, &perform)
      movements << make_movement(:charge, name, &perform)
    end
    
    def add_payment(name, &perform)
      movements << make_movement(:payment, name, &perform)
    end
    
    def add_timed_payment(name, amount, date)
      movements << make_timed_movement(name, -amount, date)
    end
    
    def add_timed_charge(name, amount, date)
      movements << make_timed_movement(name, amount, date)
    end
    
    private
    
    def make_movement(type, name, &perform)
      strategy = periodical_strategies.instance_eval(&perform)
      amount = type == :payment ? -strategy.amount : strategy.amount
      
      movement = PeriodicalMovement.new(Movement.new(type, name, amount))
      movement.can_perform_strategy = strategy.can_perform
      movement
    end
    
    def make_timed_movement(name, amount, date)
      movement = DailyRestrictedMovement.new(Movement.new(kind_for(amount), name, amount))
      movement.date = date
      movement
    end
    
    def kind_for(amount)
      amount > 0 ? :charge : :payment
    end
  end
end
